@@ -19,6 +19,7 @@ from chip.ChipDeviceCtrl import (
 from chip.clusters import Attribute, Objects as Clusters
 from chip.clusters.Attribute import ValueDecodeFailure
 from chip.clusters.ClusterObjects import ALL_ATTRIBUTES, ALL_CLUSTERS, Cluster
+from chip.discovery import CommissionableNode as CommissionableNodeData
 from chip.exceptions import ChipStackError
 
 from matter_server.server.helpers.attributes import parse_attributes_from_read_result
@@ -52,6 +53,7 @@ DATA_KEY_LAST_NODE_ID = "last_node_id"
 
 LOGGER = logging.getLogger(__name__)
 MAX_POLL_INTERVAL = 600
+MAX_COMMISSION_RETRIES = 3
 
 BASE_SUBSCRIBE_ATTRIBUTES: tuple[Attribute.AttributePath, Attribute.AttributePath] = (
     # all endpoints, BasicInformation cluster
@@ -66,7 +68,7 @@ BASE_SUBSCRIBE_ATTRIBUTES: tuple[Attribute.AttributePath, Attribute.AttributePat
     ),
 )
 
-# pylint: disable=too-many-lines,too-many-locals,too-many-statements
+# pylint: disable=too-many-lines,too-many-locals,too-many-statements,too-many-branches
 
 
 class MatterDeviceController:
@@ -91,7 +93,6 @@ class MatterDeviceController:
         self.wifi_credentials_set: bool = False
         self.thread_credentials_set: bool = False
         self.compressed_fabric_id: int | None = None
-        self._resolve_lock: asyncio.Lock = asyncio.Lock()
         self._node_lock: dict[int, asyncio.Lock] = {}
 
     async def initialize(self) -> None:
@@ -184,19 +185,31 @@ class MatterDeviceController:
 
         node_id = self._get_next_node_id()
 
-        LOGGER.info(
-            "Starting Matter commissioning with code using Node ID %s.", node_id
-        )
-        success = await self._call_sdk(
-            self.chip_controller.CommissionWithCode,
-            setupPayload=code,
-            nodeid=node_id,
-            networkOnly=network_only,
-        )
-        if not success:
-            raise NodeCommissionFailed(
-                f"Commission with code failed for node {node_id}"
+        attempts = 0
+        # we retry commissioning a few times as we've seen devices in the wild
+        # that are a bit unstable.
+        # by retrying, we increase the chances of a successful commisssion
+        while attempts <= MAX_COMMISSION_RETRIES:
+            LOGGER.info(
+                "Starting Matter commissioning with code using Node ID %s (attempt %s/%s).",
+                node_id,
+                attempts,
+                MAX_COMMISSION_RETRIES,
             )
+            success = await self._call_sdk(
+                self.chip_controller.CommissionWithCode,
+                setupPayload=code,
+                nodeid=node_id,
+                networkOnly=network_only,
+            )
+            if success:
+                break
+            if not success and attempts >= MAX_COMMISSION_RETRIES:
+                raise NodeCommissionFailed(
+                    f"Commission with code failed for node {node_id}."
+                )
+            await asyncio.sleep(5)
+            attempts += 1
         LOGGER.info("Matter commissioning of Node ID %s successful.", node_id)
 
         # perform full (first) interview of the device
@@ -246,41 +259,69 @@ class MatterDeviceController:
             raise RuntimeError("Device Controller not initialized.")
 
         node_id = self._get_next_node_id()
+        if ip_addr is not None:
+            ip_addr = self.server.scope_ipv6_lla(ip_addr)
 
-        if ip_addr is None:
-            LOGGER.info(
-                "Starting Matter commissioning on network using Node ID %s.", node_id
-            )
-            success = await self._call_sdk(
-                self.chip_controller.CommissionOnNetwork,
-                nodeId=node_id,
-                setupPinCode=setup_pin_code,
-                filterType=filter_type,
-                filter=filter,
-            )
-            if not success:
-                raise NodeCommissionFailed(
-                    f"Commission on network failed for node {node_id}"
+        attempts = 0
+        # we retry commissioning a few times as we've seen devices in the wild
+        # that are a bit unstable.
+        # by retrying, we increase the chances of a successful commisssion
+        while attempts <= MAX_COMMISSION_RETRIES:
+            if ip_addr is None:
+                # regular CommissionOnNetwork if no IP address provided
+                LOGGER.info(
+                    "Starting Matter commissioning on network using Node ID %s (attempt %s/%s).",
+                    node_id,
+                    attempts,
+                    MAX_COMMISSION_RETRIES,
                 )
-        else:
-            LOGGER.info(
-                "Starting Matter commissioning with IP using Node ID %s.", node_id
-            )
-            success = await self._call_sdk(
-                self.chip_controller.CommissionIP,
-                nodeid=node_id,
-                setupPinCode=setup_pin_code,
-                ipaddr=ip_addr,
-            )
-            if not success:
-                raise NodeCommissionFailed(
-                    f"Commission using IP failed for node {node_id}"
+                success = await self._call_sdk(
+                    self.chip_controller.CommissionOnNetwork,
+                    nodeId=node_id,
+                    setupPinCode=setup_pin_code,
+                    filterType=filter_type,
+                    filter=filter,
                 )
+            else:
+                LOGGER.info(
+                    "Starting Matter commissioning using Node ID %s and IP %s (attempt %s/%s).",
+                    node_id,
+                    ip_addr,
+                    attempts,
+                    MAX_COMMISSION_RETRIES,
+                )
+                success = await self._call_sdk(
+                    self.chip_controller.CommissionIP,
+                    nodeid=node_id,
+                    setupPinCode=setup_pin_code,
+                    ipaddr=ip_addr,
+                )
+            if success:
+                break
+            if not success and attempts >= MAX_COMMISSION_RETRIES:
+                raise NodeCommissionFailed(f"Commissioning failed for node {node_id}.")
+            await asyncio.sleep(5)
+            attempts += 1
 
         LOGGER.info("Matter commissioning of Node ID %s successful.", node_id)
 
-        # full interview of the device
-        await self.interview_node(node_id)
+        # perform full (first) interview of the device
+        # we retry the interview max 3 times as it may fail in noisy
+        # RF environments (in case of thread), mdns trouble or just flaky devices.
+        # retrying both the mdns resolve and (first) interview, increases the chances
+        # of a successful device commission.
+        retries = 3
+        while retries:
+            try:
+                await self.interview_node(node_id)
+            except NodeInterviewFailed as err:
+                if retries <= 0:
+                    raise err
+                retries -= 1
+                LOGGER.warning("Unable to interview Node %s: %s", node_id, err)
+                await asyncio.sleep(5)
+            else:
+                break
         # make sure we start a subscription for this newly added node
         await self._subscribe_node(node_id)
         LOGGER.info("Commissioning of Node ID %s completed.", node_id)
@@ -345,7 +386,7 @@ class MatterDeviceController:
     @api_command(APICommand.DISCOVER)
     async def discover_commissionable_nodes(
         self,
-    ) -> CommissionableNode | list[CommissionableNode] | None:
+    ) -> CommissionableNodeData | list[CommissionableNodeData] | None:
         """Discover Commissionable Nodes (discovered on BLE or mDNS)."""
         if self.chip_controller is None:
             raise RuntimeError("Device Controller not initialized.")
@@ -353,7 +394,17 @@ class MatterDeviceController:
         result = await self._call_sdk(
             self.chip_controller.DiscoverCommissionableNodes,
         )
-        return result
+
+        def convert(cn: CommissionableNode) -> CommissionableNodeData:
+            cnd = CommissionableNodeData()
+            # pylint: disable=no-member
+            for field in CommissionableNodeData.__dataclass_fields__:
+                setattr(cnd, field, getattr(cn, field))
+            return cnd
+
+        if isinstance(result, list):
+            return [convert(c) for c in result]
+        return convert(result)
 
     @api_command(APICommand.INTERVIEW_NODE)
     async def interview_node(self, node_id: int) -> None:
@@ -509,19 +560,25 @@ class MatterDeviceController:
                 f"Node {node_id} does not exist or has not been interviewed."
             )
 
+        LOGGER.info("Removing Node ID %s.", node_id)
+
+        # Remove and cancel any existing interview/subscription reschedule timer
+        if existing := self._sub_retry_timer.pop(node_id, None):
+            existing.cancel()
+
         # shutdown any existing subscriptions
         if sub := self._subscriptions.pop(node_id, None):
             await self._call_sdk(sub.Shutdown)
 
-        # pop any existing interview/subscription reschedule timer
-        self._sub_retry_timer.pop(node_id, None)
-
         node = self._nodes.pop(node_id)
         self.server.storage.remove(
             DATA_KEY_NODES,
             subkey=str(node_id),
         )
         self.server.storage.save(immediate=True)
+        LOGGER.info("Node ID %s successfully removed from Matter server.", node_id)
+
+        self.server.signal_event(EventType.NODE_REMOVED, node_id)
 
         assert node is not None
 
@@ -530,16 +587,29 @@ class MatterDeviceController:
             Clusters.OperationalCredentials.Attributes.CurrentFabricIndex,
         )
         fabric_index = node.attributes[attribute_path]
-
-        self.server.signal_event(EventType.NODE_REMOVED, node_id)
-
-        await self.chip_controller.SendCommand(
-            nodeid=node_id,
-            endpoint=0,
-            payload=Clusters.OperationalCredentials.Commands.RemoveFabric(
-                fabricIndex=fabric_index,
-            ),
-        )
+        result: Clusters.OperationalCredentials.Commands.NOCResponse | None = None
+        try:
+            result = await self.chip_controller.SendCommand(
+                nodeid=node_id,
+                endpoint=0,
+                payload=Clusters.OperationalCredentials.Commands.RemoveFabric(
+                    fabricIndex=fabric_index,
+                ),
+            )
+        except ChipStackError as err:
+            LOGGER.warning("Removing current fabric from device failed.", exc_info=err)
+            return
+        if (
+            result is None
+            or result.statusCode
+            == Clusters.OperationalCredentials.Enums.NodeOperationalCertStatusEnum.kOk
+        ):
+            LOGGER.info("Successfully removed Home Assistant fabric from device.")
+        else:
+            LOGGER.warning(
+                "Removing current fabric from device failed with status code %d.",
+                result.statusCode,
+            )
 
     @api_command(APICommand.SUBSCRIBE_ATTRIBUTE)
     async def subscribe_attribute(
@@ -938,22 +1008,19 @@ class MatterDeviceController:
         if self.chip_controller is None:
             raise RuntimeError("Device Controller not initialized.")
         try:
-            # the sdk crashes when multiple resolves happen at the same time
-            # guard simultane resolves with a lock.
-            async with self._resolve_lock:
-                LOGGER.log(
-                    log_level,
-                    "Attempting to resolve node %s... (attempt %s of %s)",
-                    node_id,
-                    attempt,
-                    retries,
-                )
-                return await self._call_sdk(
-                    self.chip_controller.GetConnectedDeviceSync,
-                    nodeid=node_id,
-                    allowPASE=False,
-                    timeoutMs=None,
-                )
+            LOGGER.log(
+                log_level,
+                "Attempting to resolve node %s... (attempt %s of %s)",
+                node_id,
+                attempt,
+                retries,
+            )
+            return await self._call_sdk(
+                self.chip_controller.GetConnectedDeviceSync,
+                nodeid=node_id,
+                allowPASE=False,
+                timeoutMs=None,
+            )
         except (ChipStackError, TimeoutError) as err:
             if attempt >= retries:
                 # when we're out of retries, raise NodeNotResolving
